#!/usr/bin/env python3
import asyncio
import logging
import sys
from time import sleep

from aiohttp import ClientConnectionError, ServerDisconnectedError
from apscheduler.schedulers import SchedulerAlreadyRunningError
from nio import (
    AsyncClient,
    AsyncClientConfig,
    InviteMemberEvent,
    LocalProtocolError,
    LoginError,
    MegolmEvent,
    RoomMessageText,
)

from matrix_reminder_bot.callbacks import Callbacks
from matrix_reminder_bot.config import CONFIG
from matrix_reminder_bot.reminder import SCHEDULER
from matrix_reminder_bot.storage import Storage

logger = logging.getLogger(__name__)


async def main():
    # Read config file
    # A different config file path can be specified as the first command line arg
    if len(sys.argv) > 1:
        config_filepath = sys.argv[1]
    else:
        config_filepath = "config.yaml"
    CONFIG.read_config(config_filepath)

    # Configure the python job scheduler
    SCHEDULER.configure({"apscheduler.timezone": CONFIG.timezone})

    # Configuration options for the AsyncClient
    client_config = AsyncClientConfig(
        max_limit_exceeded=0,
        max_timeouts=0,
        store_sync_tokens=True,
        encryption_enabled=True,
    )

    # Initialize the matrix client
    client = AsyncClient(
        CONFIG.homeserver_url,
        CONFIG.user_id,
        device_id=CONFIG.device_id,
        store_path=CONFIG.store_path,
        config=client_config,
    )

    # Configure the database
    store = Storage(client)

    # Set up event callbacks
    callbacks = Callbacks(client, store)
    client.add_event_callback(callbacks.message, (RoomMessageText,))
    client.add_event_callback(callbacks.invite, (InviteMemberEvent,))
    client.add_event_callback(callbacks.decryption_failure, (MegolmEvent,))

    # Keep trying to reconnect on failure (with some time in-between)
    while True:
        try:
            # Try to login with the configured username/password
            try:
                login_response = await client.login(
                    password=CONFIG.user_password, device_name=CONFIG.device_name,
                )

                # Check if login failed. Usually incorrect password
                if type(login_response) == LoginError:
                    logger.error("Failed to login: %s", login_response.message)
                    logger.warning("Trying again in 15s...")

                    # Sleep so we don't bombard the server with login requests
                    sleep(15)
                    continue
            except LocalProtocolError as e:
                # There's an edge case here where the user hasn't installed the correct C
                # dependencies. In that case, a LocalProtocolError is raised on login.
                logger.fatal(
                    "Failed to login. Have you installed the correct dependencies? "
                    "https://github.com/poljar/matrix-nio#installation "
                    "Error: %s",
                    e,
                )
                return False

            # Login succeeded!

            logger.info(f"Logged in as {CONFIG.user_id}")
            logger.info("Startup complete")

            # Allow jobs to fire
            try:
                SCHEDULER.start()
            except SchedulerAlreadyRunningError:
                pass

            await client.sync_forever(timeout=30000, full_state=True)

        except (ClientConnectionError, ServerDisconnectedError, TimeoutError):
            logger.warning("Unable to connect to homeserver, retrying in 15s...")

            # Sleep so we don't bombard the server with login requests
            sleep(15)
        except Exception:
            logger.exception("Unknown exception occurred:")
            logger.warning("Restarting in 15s...")

            # Sleep so we don't bombard the server with login requests
            sleep(15)
        finally:
            # Make sure to close the client connection on disconnect
            await client.close()


if __name__ == "__main__":
    asyncio.get_event_loop().run_until_complete(main())



class ConfigError(RuntimeError):
    """An error encountered during reading the config file

    Args:
        msg (str): The message displayed to the bot operator on error
    """

    def __init__(self, msg: str):
        super().__init__("%s" % (msg,))


class CommandError(RuntimeError):
    """An error encountered while processing a command

    Args:
        msg: The message that is sent back to the user on error
    """

    def __init__(self, msg: str):
        super().__init__("%s" % (msg,))
        self.msg = msg


class CommandSyntaxError(RuntimeError):
    """An error encountered if syntax of a called command was violated"""

    def __init__(self):
        super().__init__()



import logging

from nio import (
    AsyncClient,
    InviteMemberEvent,
    JoinError,
    MatrixRoom,
    MegolmEvent,
    RoomMessageText,
)

from matrix_reminder_bot.bot_commands import Command
from matrix_reminder_bot.config import CONFIG
from matrix_reminder_bot.errors import CommandError
from matrix_reminder_bot.functions import send_text_to_room
from matrix_reminder_bot.storage import Storage

logger = logging.getLogger(__name__)


class Callbacks(object):
    """Callback methods that fire on certain matrix events

    Args:
        client: nio client used to interact with matrix
        store: Bot storage
    """

    def __init__(self, client: AsyncClient, store: Storage):
        self.client = client
        self.store = store

    async def message(self, room: MatrixRoom, event: RoomMessageText):
        """Callback for when a message event is received"""
        # Extract the formatted message text, or the basic text if no formatting is available
        msg = event.formatted_body or event.body
        if not msg:
            return

        # Ignore messages from ourselves
        if event.sender == self.client.user:
            return

        # Check whether this is a command
        #
        # We use event.body here as formatted bodies can start with <p> instead of the
        # command prefix
        if not event.body.startswith(CONFIG.command_prefix):
            return

        # It's possible to erraneous HTML tags to appear before the command prefix when
        # making use of event.formatted_body. Typically <p> when commands include newlines
        # This can cause the command code to break.
        #
        # We've already determined this is a command, so strip everything in the msg before
        # the first instance of the command prefix as a workaround
        prefix_index = msg.find(CONFIG.command_prefix)
        if prefix_index != -1:
            msg = msg[prefix_index:]
        else:
            # This formatted body doesn't contain the command prefix for some reason.
            # Use event.body instead then
            msg = event.body

        logger.debug("Command received: %s", msg)

        # Assume this is a command and attempt to process
        command = Command(self.client, self.store, msg, room, event)

        try:
            await command.process()
        except CommandError as e:
            # An expected error occurred. Inform the user
            msg = f"Error: {e.msg}"
            await send_text_to_room(self.client, room.room_id, msg)

            # Print traceback
            logger.exception("CommandError while processing command:")
        except Exception as e:
            # An unknown error occurred. Inform the user
            msg = f"An unknown error occurred: {e}"
            await send_text_to_room(self.client, room.room_id, msg)

            # Print traceback
            logger.exception("Unknown error while processing command:")

    async def invite(self, room: MatrixRoom, event: InviteMemberEvent):
        """Callback for when an invite is received. Join the room specified in the invite"""
        logger.debug(f"Got invite to {room.room_id} from {event.sender}.")

        # Attempt to join 3 times before giving up
        for attempt in range(3):
            result = await self.client.join(room.room_id)
            if type(result) == JoinError:
                logger.error(
                    f"Error joining room {room.room_id} (attempt %d): %s",
                    attempt,
                    result.message,
                )
            else:
                logger.info(f"Joined {room.room_id}")
                break

    async def decryption_failure(self, room: MatrixRoom, event: MegolmEvent):
        """Callback for when an event fails to decrypt. Inform the user"""
        logger.error(
            f"Failed to decrypt event '{event.event_id}' in room '{room.room_id}'!"
            f"\n\n"
            f"Tip: try using a different device ID in your config file and restart."
            f"\n\n"
            f"If all else fails, delete your store directory and let the bot recreate "
            f"it (your reminders will NOT be deleted, but the bot may respond to existing "
            f"commands a second time)."
        )

        user_msg = (
            "Unable to decrypt this message. "
            "Check whether you've chosen to only encrypt to trusted devices."
        )

        await send_text_to_room(
            self.client, room.room_id, user_msg, reply_to_event_id=event.event_id,
        )



import logging
from datetime import datetime, timedelta, timezone
from typing import Optional, Tuple

import dateparser
import pytz
from apscheduler.job import Job
from nio import AsyncClient, MatrixRoom
from nio.events.room_events import RoomMessageText
from readabledelta import readabledelta

from matrix_reminder_bot.config import CONFIG
from matrix_reminder_bot.errors import CommandError, CommandSyntaxError
from matrix_reminder_bot.functions import command_syntax, send_text_to_room
from matrix_reminder_bot.reminder import ALARMS, REMINDERS, SCHEDULER, Reminder
from matrix_reminder_bot.storage import Storage

logger = logging.getLogger(__name__)


class Command(object):
    def __init__(
        self,
        client: AsyncClient,
        store: Storage,
        command: str,
        room: MatrixRoom,
        event: RoomMessageText,
    ):
        """A command made by a user

        Args:
            client: The client to communicate to matrix with
            store: Bot storage
            command: The command and arguments
            room: The room the command was sent in
            event: The event describing the command
        """
        self.client = client
        self.store = store
        self.room = room
        self.event = event

        msg_without_prefix = command[
            len(CONFIG.command_prefix) :
        ]  # Remove the cmd prefix
        self.args = (
            msg_without_prefix.split()
        )  # Get a list of all items, split by spaces
        self.command = self.args.pop(
            0
        )  # Remove the first item and save as the command (ex. `remindme`)

    def _parse_reminder_command_args_for_cron(self) -> Tuple[str, str]:
        """Processes the list of arguments when a cron tab is present

        Returns:
            A tuple containing the cron tab and the reminder text.
        """

        # Retrieve the cron tab and reminder text

        # Remove "cron" from the argument list
        args = self.args[1:]

        # Combine arguments into a string
        args_str = " ".join(args)
        logger.debug("Parsing cron command arguments: %s", args_str)

        # Split into cron tab and reminder text
        cron_tab, reminder_text = args_str.split(";", maxsplit=1)

        return cron_tab, reminder_text.strip()

    def _parse_reminder_command_args(self) -> Tuple[datetime, str, Optional[timedelta]]:
        """Processes the list of arguments and returns parsed reminder information

        Returns:
            A tuple containing the start time of the reminder as a datetime, the reminder text,
            and a timedelta representing how often to repeat the reminder, or None depending on
            whether this is a recurring reminder.

        Raises:
            CommandError: if a time specified in the user command is invalid or in the past
        """
        args_str = " ".join(self.args)
        logger.debug("Parsing command arguments: %s", args_str)

        try:
            time_str, reminder_text = args_str.split(";", maxsplit=1)
        except ValueError:
            raise CommandSyntaxError()
        logger.debug("Got time: %s", time_str)

        # Clean up the input
        time_str = time_str.strip().lower()
        reminder_text = reminder_text.strip()

        # Determine whether this is a recurring command
        # Recurring commands take the form:
        # every <recurse time>, <start time>, <text>
        recurring = time_str.startswith("every")
        recurse_timedelta = None
        if recurring:
            # Remove "every" and retrieve the recurse time
            recurse_time_str = time_str[len("every") :].strip()
            logger.debug("Got recurring time: %s", recurse_time_str)

            # Convert the recurse time to a datetime object
            recurse_time = self._parse_str_to_time(recurse_time_str)

            # Generate a timedelta between now and the recurring time
            # `recurse_time` is guaranteed to always be in the future
            current_time = self._get_datetime_now(CONFIG.timezone)

            recurse_timedelta = recurse_time - current_time
            logger.debug("Recurring timedelta: %s", recurse_timedelta)

            # Extract the start time
            time_str, reminder_text = reminder_text.split(";", maxsplit=1)
            reminder_text = reminder_text.strip()

            logger.debug("Start time: %s", time_str)

        # Convert start time string to a datetime object
        time = self._parse_str_to_time(time_str)

        return time, reminder_text, recurse_timedelta

    def _parse_str_to_time(self, time_str: str) -> datetime:
        """Converts a human-readable, future time string to a datetime object

        Args:
            time_str: The time to convert

        Returns:
            datetime: A datetime if conversion was successful

        Raises:
            CommandError: if conversion was not successful, or time is in the past.
        """
        time = dateparser.parse(
            time_str,
            settings={
                "PREFER_DATES_FROM": "future",
                "TIMEZONE": CONFIG.timezone,
                "RETURN_AS_TIMEZONE_AWARE": True,
            },
        )
        if not time:
            raise CommandError(f"The given time '{time_str}' is invalid.")

        # Disallow times in the past
        if time < self._get_datetime_now(CONFIG.timezone):
            raise CommandError(f"The given time '{time_str}' is in the past.")

        # Round datetime object to the nearest second for nicer display
        time = time.replace(microsecond=0)

        return time

    def _get_datetime_now(self, tz: str) -> datetime:
        """Returns a timezone-aware datetime object of the current time"""
        # Get a datetime with no timezone information
        no_timezone_datetime = datetime(2009, 9, 1)

        # Create a datetime.timezone object with the correct offset from UTC
        offset = timezone(pytz.timezone(tz).utcoffset(no_timezone_datetime))

        # Get datetime.now with that offset
        now = datetime.now(offset)

        # Round to the nearest second for nicer display
        return now.replace(microsecond=0)

    async def _confirm_reminder(self, reminder: Reminder):
        """Sends a message to the room confirming the reminder is set

        Args:
            reminder: The Reminder to confirm
        """
        if reminder.cron_tab:
            # Special-case cron-style reminders. We currently don't do any special
            # parsing for them
            await send_text_to_room(
                self.client, self.room.room_id, "OK, I will remind you!"
            )

            return

        # Convert a timedelta to a formatted time (ex. May 25 2020, 01:31)
        start_time = reminder.start_time.replace(
            tzinfo=pytz.timezone(reminder.timezone)
        )
        human_readable_start_time = start_time.strftime("%b %d %Y, %H:%M")

        # Get a textual representation of who will be notified by this reminder
        target = "you" if reminder.target_user else "everyone in the room"

        # Build the response string
        text = f"OK, I will remind {target} on {human_readable_start_time}"

        if reminder.recurse_timedelta:
            # Inform the user how often their reminder will repeat
            text += f", and again every {readabledelta(reminder.recurse_timedelta)}"

        # Add some punctuation
        text += "!"

        if reminder.alarm:
            # Inform the user that an alarm is attached to this reminder
            text += (
                f"\n\nWhen this reminder goes off, an alarm will sound every "
                f"5 minutes until silenced. Alarms may be silenced using the "
                f"`{CONFIG.command_prefix}silence` command."
            )

        # Send the message to the room
        await send_text_to_room(self.client, self.room.room_id, text)

    async def _remind(self, target: Optional[str] = None, alarm: bool = False):
        """Create a reminder or an alarm with a given target

        Args:
            target: A user ID if this reminder will mention a single user. If None,
                the reminder will mention the whole room
            alarm: Whether this reminder is an alarm. It will fire every 5m after it
                normally fires until silenced.
        """
        # Check whether the time is in human-readable format ("tomorrow at 5pm") or cron-tab
        # format ("* * * * 2,3,4 *"). We differentiate by checking if the time string starts
        # with "cron"
        cron_tab = None
        start_time = None
        recurse_timedelta = None
        if " ".join(self.args).lower().startswith("cron"):
            cron_tab, reminder_text = self._parse_reminder_command_args_for_cron()

            logger.debug(
                "Creating reminder in room %s with cron tab %s: %s",
                self.room.room_id,
                cron_tab,
                reminder_text,
            )
        else:
            (
                start_time,
                reminder_text,
                recurse_timedelta,
            ) = self._parse_reminder_command_args()

            logger.debug(
                "Creating reminder in room %s with delta %s: %s",
                self.room.room_id,
                recurse_timedelta,
                reminder_text,
            )

        if (self.room.room_id, reminder_text.upper()) in REMINDERS:
            await send_text_to_room(
                self.client,
                self.room.room_id,
                "A similar reminder already exists. Please delete that one first.",
            )
            return

        # Create the reminder
        reminder = Reminder(
            self.client,
            self.store,
            self.room.room_id,
            reminder_text,
            start_time=start_time,
            timezone=CONFIG.timezone,
            cron_tab=cron_tab,
            recurse_timedelta=recurse_timedelta,
            target_user=target,
            alarm=alarm,
        )

        # Record the reminder
        REMINDERS[(self.room.room_id, reminder_text.upper())] = reminder
        self.store.store_reminder(reminder)

        # Send a message to the room confirming the creation of the reminder
        await self._confirm_reminder(reminder)

    async def process(self):
        """Process the command"""
        if self.command == "remindme":
            await self._remind_me()
        elif self.command == "remindroom":
            await self._remind_room()
        elif self.command == "alarmme":
            await self._alarm_me()
        elif self.command == "alarmroom":
            await self._alarm_room()
        elif self.command in ["listreminders", "listalarms"]:
            await self._list_reminders()
        elif self.command in [
            "delreminder",
            "deletereminder",
            "removereminder",
            "cancelreminder",
            "delalarm",
            "deletealarm",
            "removealarm",
            "cancelalarm",
        ]:
            await self._delete_reminder()
        elif self.command == "silence":
            await self._silence()
        elif self.command == "help":
            await self._help()

    @command_syntax("[every <recurring time>;] <start time>; <reminder text>")
    async def _remind_me(self):
        """Set a reminder that will remind only the user who created it"""
        await self._remind(target=self.event.sender)

    @command_syntax("[every <recurring time>;] <start time>; <reminder text>")
    async def _remind_room(self):
        """Set a reminder that will mention the room that the reminder was created in"""
        await self._remind()

    @command_syntax("[every <recurring time>;] <start time>; <reminder text>")
    async def _alarm_me(self):
        """Set a reminder with an alarm that will remind only the user who created it"""
        await self._remind(target=self.event.sender, alarm=True)

    @command_syntax("[every <recurring time>;] <start time>; <reminder text>")
    async def _alarm_room(self):
        """Set a reminder with an alarm that when fired will mention the room that the
        reminder was created in
        """
        await self._remind(alarm=True)

    @command_syntax("[<reminder text>]")
    async def _silence(self):
        """Silences an ongoing alarm"""
        alarm_job = None

        # Attempt to find a reminder with an alarm currently going off
        reminder_text = " ".join(self.args)
        if reminder_text:
            # Find the alarm job via its reminder text
            alarm_job = ALARMS.get((self.room.room_id, reminder_text.upper()))

            if alarm_job:
                await self._remove_and_silence_alarm(alarm_job, reminder_text)
                text = f"Alarm '{reminder_text}' silenced."
            else:
                # We didn't find an alarm with that reminder text
                #
                # Be helpful and check if this is a known reminder without an alarm
                # currently going off
                reminder = REMINDERS.get((self.room.room_id, reminder_text.upper()))
                if reminder:
                    text = (
                        f"The reminder '{reminder_text}' does not currently have an "
                        f"alarm going off."
                    )
                else:
                    # Nope, can't find it
                    text = f"Unknown alarm or reminder '{reminder_text}'."

        else:
            # No reminder text provided. Check if there's a reminder currently firing
            # in the room instead then
            for alarm_info, job in ALARMS.items():
                if alarm_info[0] == self.room.room_id:
                    # Found one!
                    reminder_text = alarm_info[
                        1
                    ].capitalize()  # normalize the text a bit

                    await self._remove_and_silence_alarm(job, reminder_text)
                    text = f"Alarm '{reminder_text}' silenced."

                    # Prevent the `else` clause from being triggered
                    break
            else:
                # If we didn't find any alarms...
                text = "No alarms are currently firing in this room."

        await send_text_to_room(self.client, self.room.room_id, text)

    async def _remove_and_silence_alarm(self, alarm_job: Job, reminder_text: str):
        # We found a reminder with an alarm. Remove it from the dict of current
        # alarms
        ALARMS.pop((self.room.room_id, reminder_text.upper()), None)

        if SCHEDULER.get_job(alarm_job.id):
            # Silence the alarm job
            alarm_job.remove()

    @command_syntax("")
    async def _list_reminders(self):
        """Format and show known reminders for the current room

        Sends a message listing them in the following format:

            Reminders for this room:

            <start time>: <reminder text> [(every <recurring time>)] [(has alarm)]

        or if there are no reminders set:

            There are no reminders for this room.
        """
        reminders = []
        for reminder in REMINDERS.values():
            # Filter out reminders that don't belong to this room
            if reminder.room_id != self.room.room_id:
                continue

            # If this is a cron-style reminder, just print the cron tab
            if reminder.cron_tab:
                line = f"`{reminder.cron_tab}`: {reminder.reminder_text}"
                reminders.append(line)
                continue

            # Format the start time into something readable
            human_readable_start_time = reminder.start_time.strftime("%b %d %Y, %H:%M")

            line = f"{human_readable_start_time}: {reminder.reminder_text}"

            # Display a recurring time if available
            if reminder.recurse_timedelta:
                # Get a nice, human-readable version to print
                line += f" (every {readabledelta(reminder.recurse_timedelta)})"

            # Note that an alarm exists if available
            if reminder.alarm:
                line += " (has alarm)"

            reminders.append(line)

        if reminders:
            text = "Reminders for this room:\n\n"
            text += "\n\n".join(reminders)
        else:
            text = "There are no reminders for this room."

        await send_text_to_room(self.client, self.room.room_id, text)

    @command_syntax("<reminder text>")
    async def _delete_reminder(self):
        """Delete a reminder via its reminder text"""
        reminder_text = " ".join(self.args)
        if not reminder_text:
            raise CommandSyntaxError()

        logger.debug("Known reminders: %s", REMINDERS)
        logger.debug(
            "Deleting reminder in room %s: %s", self.room.room_id, reminder_text
        )

        reminder = REMINDERS.get((self.room.room_id, reminder_text.upper()))
        if reminder:
            # Cancel the reminder and associated alarms
            reminder.cancel()

            text = "Reminder cancelled."
        else:
            text = f"Unknown reminder '{reminder_text}'."

        await send_text_to_room(self.client, self.room.room_id, text)

    @command_syntax("")
    async def _help(self):
        """Show the help text"""
        # Ensure we don't tell the user to use something other than their configured command
        # prefix
        c = CONFIG.command_prefix

        if not self.args:
            text = (
                f"Hello, I am a reminder bot! Use `{c}help reminders` "
                f"to view available commands."
            )
            await send_text_to_room(self.client, self.room.room_id, text)
            return

        topic = self.args[0]

        # Simply way to check for plurals
        if topic.startswith("reminder"):
            text = f"""
**Reminders**

Create an optionally recurring reminder that notifies the reminder creator:

```
{c}remindme [every <recurring time>;] <start time>; <reminder text>
```

Create an optionally recurring reminder that notifies the whole room.
(Note that the bot will need appropriate permissions to mention
the room):

```
{c}remindroom [every <recurring time>;] <start time>; <reminder text>
```

List all active reminders for a room:

```
{c}listreminders
```

Cancel a reminder:

```
{c}cancelreminder <reminder text>
```

**Alarms**

Create a reminder that will repeatedly sound every 5m after its usual
fire time. Otherwise, the syntax is the same as a reminder:

```
{c}alarmme [every <recurring time>;] <start time>; <reminder text>
```

or for notifying the whole room:

```
{c}alarmroom [every <recurring time>;] <start time>; <reminder text>
```

Once firing, an alarm can be silenced with:

```
{c}silence [<reminder text>]
```

**Cron-tab Syntax**

If you need more complicated recurring reminders, you can make use of
cron-tab syntax:

```
{c}remindme cron <min> <hour> <day of month> <month> <day of week>; <reminder text>
```

This syntax is supported by any `{c}remind...` or `{c}alarm...` command above.
"""
        else:
            # Unknown help topic
            return

        await send_text_to_room(self.client, self.room.room_id, text)

    async def _unknown_command(self):
        """Computer says 'no'."""
        await send_text_to_room(
            self.client,
            self.room.room_id,
            f"Unknown help topic '{self.command}'. Try the 'help' command for more "
            f"information.",
        )



import sys

# Check that we're not running on an unsupported Python version.
if sys.version_info < (3, 6):
    print("matrix_reminder_bot requires Python 3.6 or above.")
    sys.exit(1)

__version__ = "0.2.0"



import logging
from datetime import datetime, timedelta
from typing import Dict, Optional, Tuple

from apscheduler.job import Job
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger
from apscheduler.util import timedelta_seconds
from nio import AsyncClient

from matrix_reminder_bot.config import CONFIG
from matrix_reminder_bot.functions import send_text_to_room

logger = logging.getLogger(__name__)

# The object that runs callbacks at a certain time
SCHEDULER = AsyncIOScheduler()

# How often an alarm should sound after the reminder it's attached to
ALARM_TIMEDELTA = timedelta(minutes=5)


class Reminder(object):
    """An object containing information about a reminder, when it should go off,
    whether it is recurring, etc.

    Args:
        client: The matrix client
        store: A Storage object
        room_id: The ID of the room the reminder should appear in
        start_time: When the reminder should first go off
        timezone: The database name of the timezone this reminder should act within
        reminder_text: The text to include in the reminder message
        recurse_timedelta: Optional. How often to repeat the reminder
        target_user: Optional. A user ID of a specific user to mention in the room while
            reminding
        alarm: Whether this reminder is an alarm. Alarms are reminders that fire every 5m
            after they go off normally, until they are silenced.
    """

    def __init__(
        self,
        client: AsyncClient,
        store,
        room_id: str,
        reminder_text: str,
        start_time: Optional[datetime] = None,
        timezone: Optional[str] = None,
        recurse_timedelta: Optional[timedelta] = None,
        cron_tab: Optional[str] = None,
        target_user: Optional[str] = None,
        alarm: bool = False,
    ):
        self.client = client
        self.store = store
        self.room_id = room_id
        self.timezone = timezone
        self.start_time = start_time
        self.reminder_text = reminder_text
        self.cron_tab = cron_tab
        self.recurse_timedelta = recurse_timedelta
        self.target_user = target_user
        self.alarm = alarm

        # Schedule the reminder

        # Determine how the reminder is triggered
        if cron_tab:
            # Set up a cron trigger
            trigger = CronTrigger.from_crontab(cron_tab, timezone=timezone)
        elif recurse_timedelta:
            # Use an interval trigger (runs multiple times)
            trigger = IntervalTrigger(
                # timedelta.seconds does NOT give you the timedelta converted to seconds
                # Use a method from apscheduler instead
                seconds=int(timedelta_seconds(recurse_timedelta)),
                start_date=start_time,
                timezone=timezone,
            )
        else:
            # Use a date trigger (runs only once)
            trigger = DateTrigger(run_date=start_time, timezone=timezone)

        # Note down the job for later manipulation
        self.job = SCHEDULER.add_job(self._fire, trigger=trigger)

        self.alarm_job = None

    async def _fire(self):
        """Called when a reminder fires"""
        logger.debug("Reminder in room %s fired: %s", self.room_id, self.reminder_text)

        # Build the reminder message
        target = self.target_user if self.target_user else "@room"
        message = f"{target} {self.reminder_text}"

        # If this reminder has an alarm attached...
        if self.alarm:
            # Inform the user that an alarm will go off
            message += (
                f"\n\n(This reminder has an alarm. You will be reminded again in 5m. "
                f"Use the `{CONFIG.command_prefix}silence` command to stop)."
            )

            # Check that an alarm is not already ongoing from a previous run
            if not (self.room_id, self.reminder_text.upper()) in ALARMS:
                # Start alarming
                self.alarm_job = SCHEDULER.add_job(
                    self._fire_alarm,
                    trigger=IntervalTrigger(
                        # timedelta.seconds does NOT give you the timedelta converted to
                        # seconds. Use a method from apscheduler instead
                        seconds=int(timedelta_seconds(ALARM_TIMEDELTA)),
                    ),
                )
                ALARMS[(self.room_id, self.reminder_text.upper())] = self.alarm_job

        # Send the message to the room
        await send_text_to_room(self.client, self.room_id, message, notice=False)

        # If this was a one-time reminder, cancel and remove from the reminders dict
        if not self.recurse_timedelta and not self.cron_tab:
            # We set cancel_alarm to False here else the associated alarms wouldn't even
            # fire
            self.cancel(cancel_alarm=False)

    async def _fire_alarm(self):
        logger.debug("Alarm in room %s fired: %s", self.room_id, self.reminder_text)

        # Build the alarm message
        target = self.target_user if self.target_user else "@room"
        message = (
            f"Alarm: {target} {self.reminder_text} "
            f"(Use `{CONFIG.command_prefix}silence [reminder text]` to silence)."
        )

        # Send the message to the room
        await send_text_to_room(self.client, self.room_id, message, notice=False)

    def cancel(self, cancel_alarm: bool = True):
        """Cancels a reminder and all recurring instances

        Args:
            cancel_alarm: Whether to also cancel alarms of this reminder
        """
        logger.debug(
            "Cancelling reminder in room %s: %s", self.room_id, self.reminder_text
        )

        # Remove from the in-memory reminder and alarm dicts
        REMINDERS.pop((self.room_id, self.reminder_text.upper()), None)

        # Delete the reminder from the database
        self.store.delete_reminder(self.room_id, self.reminder_text)

        # Delete any ongoing jobs
        if self.job and SCHEDULER.get_job(self.job.id):
            self.job.remove()

        # Cancel alarms of this reminder if required
        if cancel_alarm:
            ALARMS.pop((self.room_id, self.reminder_text.upper()), None)

            if self.alarm_job and SCHEDULER.get_job(self.alarm_job.id):
                self.alarm_job.remove()


# Global dictionaries
#
# Both feature (room_id, reminder_text) tuples as keys
#
# reminder_text should be accessed and stored as uppercase in order to
# allow for case-insensitive matching when carrying out user actions
REMINDERS: Dict[Tuple[str, str], Reminder] = {}
ALARMS: Dict[Tuple[str, str], Job] = {}



import logging
from typing import Callable, Optional

from markdown import markdown
from nio import AsyncClient, SendRetryError

from matrix_reminder_bot.config import CONFIG
from matrix_reminder_bot.errors import CommandSyntaxError

logger = logging.getLogger(__name__)


async def send_text_to_room(
    client: AsyncClient,
    room_id: str,
    message: str,
    notice: bool = True,
    markdown_convert: bool = True,
    reply_to_event_id: Optional[str] = None,
):
    """Send text to a matrix room.

    Args:
        client: The client to communicate to matrix with.

        room_id: The ID of the room to send the message to.

        message: The message content.

        notice: Whether the message should be sent with an "m.notice" message type
            (will not ping users).

        markdown_convert: Whether to convert the message content to markdown.
            Defaults to true.

        reply_to_event_id: Whether this message is a reply to another event. The event
            ID this is message is a reply to.
    """
    # Determine whether to ping room members or not
    msgtype = "m.notice" if notice else "m.text"

    content = {
        "msgtype": msgtype,
        "format": "org.matrix.custom.html",
        "body": message,
    }

    if markdown_convert:
        content["formatted_body"] = markdown(message)

    if reply_to_event_id:
        content["m.relates_to"] = {"m.in_reply_to": {"event_id": reply_to_event_id}}

    try:
        await client.room_send(
            room_id, "m.room.message", content, ignore_unverified_devices=True,
        )
    except SendRetryError:
        logger.exception(f"Unable to send message response to {room_id}")


def command_syntax(syntax: str):
    """Defines the syntax for a function, and informs the user if it is violated

    This function is intended to be used as a decorator, allowing command-handler
    functions to define the syntax that the user is supposed to use for the
    command arguments.

    The command function, passed to `outer`, can signal that this syntax has been
    violated by raising a CommandSyntaxError exception. This will then catch that
    exception and inform the user of the correct syntax for that command.

    Args:
        syntax: The syntax for the command that the user should follow
    """

    def outer(command_func: Callable):
        async def inner(self, *args, **kwargs):
            try:
                # Attempt to execute the command function
                await command_func(self, *args, **kwargs)
            except CommandSyntaxError:
                # The function indicated that there was a command syntax error
                # Inform the user of the correct syntax
                #
                # Grab the bot's configured command prefix, and the current
                # command's name from the `self` object passed to the command
                text = (
                    f"Invalid syntax. Please use "
                    f"`{CONFIG.command_prefix}{self.command} {syntax}`."
                )
                await send_text_to_room(self.client, self.room.room_id, text)

        return inner

    return outer



import logging
import os
import re
import sys
from typing import Any, List

import yaml

from matrix_reminder_bot.errors import ConfigError

logger = logging.getLogger()
logging.getLogger("peewee").setLevel(
    logging.INFO
)  # Prevent debug messages from peewee lib


class DatabaseConfig:
    def __init__(self):
        # The type of database. Supported types are 'sqlite' and 'postgres'
        self.type: str = ""
        self.connection_string: str = ""


class Config:
    def __init__(self):
        """
        Args:
            filepath: Path to the config file
        """
        # TODO: Add some comments for each of these
        # TODO: Also ensure that this commit diff is sane. Did I replace config everywhere?
        self.database: DatabaseConfig = DatabaseConfig()
        self.store_path: str = ""

        self.user_id: str = ""
        self.user_password: str = ""
        self.device_id: str = ""
        self.device_name: str = ""
        self.homeserver_url: str = ""

        self.command_prefix: str = ""

        self.timezone: str = ""

    def read_config(self, filepath: str):
        if not os.path.isfile(filepath):
            raise ConfigError(f"Config file '{filepath}' does not exist")

        # Load in the config file at the given filepath
        with open(filepath) as file_stream:
            self.config = yaml.safe_load(file_stream.read())

        # Logging setup
        formatter = logging.Formatter(
            "%(asctime)s | %(name)s [%(levelname)s] %(message)s"
        )

        log_level = self._get_cfg(["logging", "level"], default="INFO")
        logger.setLevel(log_level)

        file_logging_enabled = self._get_cfg(
            ["logging", "file_logging", "enabled"], default=False
        )
        file_logging_filepath = self._get_cfg(
            ["logging", "file_logging", "filepath"], default="bot.log"
        )
        if file_logging_enabled:
            handler = logging.FileHandler(file_logging_filepath)
            handler.setFormatter(formatter)
            logger.addHandler(handler)

        console_logging_enabled = self._get_cfg(
            ["logging", "console_logging", "enabled"], default=True
        )
        if console_logging_enabled:
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(formatter)
            logger.addHandler(handler)

        # Storage setup
        database_path = self._get_cfg(["storage", "database"], required=True)

        # We support both SQLite and Postgres backends
        # Determine which one the user intends
        sqlite_scheme = "sqlite://"
        postgres_scheme = "postgres://"
        if database_path.startswith(sqlite_scheme):
            self.database.type = "sqlite"
            self.database.connection_string = database_path[len(sqlite_scheme) :]
        elif database_path.startswith(postgres_scheme):
            self.database.type = "postgres"
            self.database.connection_string = database_path
        else:
            raise ConfigError("Invalid connection string for storage.database")

        self.store_path = self._get_cfg(["storage", "store_path"], default="store")

        # Create the store folder if it doesn't exist
        if not os.path.isdir(self.store_path):
            if not os.path.exists(self.store_path):
                os.mkdir(self.store_path)
            else:
                raise ConfigError(
                    f"storage.store_path '{self.store_path}' is not a directory"
                )

        # Matrix bot account setup
        user_id = self._get_cfg(["matrix", "user_id"], required=True)
        if not re.match("@.*:.*", user_id):
            raise ConfigError("matrix.user_id must be in the form @name:domain")
        self.user_id = user_id

        self.user_password = self._get_cfg(["matrix", "user_password"], required=True)
        self.device_id = self._get_cfg(["matrix", "device_id"], required=True)
        self.device_name = self._get_cfg(
            ["matrix", "device_name"], default="nio-template"
        )
        self.homeserver_url = self._get_cfg(["matrix", "homeserver_url"], required=True)

        self.command_prefix = self._get_cfg(["command_prefix"], default="!")

        # Reminder configuration
        self.timezone = self._get_cfg(["reminders", "timezone"], default="Etc/UTC")

    def _get_cfg(
        self, path: List[str], default: Any = None, required: bool = True,
    ) -> Any:
        """Get a config option from a path and option name, specifying whether it is
        required.

        Raises:
            ConfigError: If required is specified and the object is not found
                (and there is no default value provided), this error will be raised
        """
        # Sift through the the config until we reach our option
        config = self.config
        for name in path:
            config = config.get(name)

            # If at any point we don't get our expected option...
            if config is None:
                # Raise an error if it was required
                if required and not default:
                    raise ConfigError(f"Config option {'.'.join(path)} is required")

                # or return the default value
                return default

        # We found the option. Return it
        return config


CONFIG: Config = Config()



#!/usr/bin/env python3
import os

from setuptools import find_packages, setup


def exec_file(path_segments):
    """Execute a single python file to get the variables defined in it"""
    result = {}
    code = read_file(path_segments)
    exec(code, result)
    return result


def read_file(path_segments):
    """Read a file from the package. Takes a list of strings to join to
    make the path"""
    file_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), *path_segments)
    with open(file_path) as f:
        return f.read()


version = exec_file(("matrix_reminder_bot", "__init__.py"))["__version__"]
long_description = read_file(("README.md",))


setup(
    name="matrix-reminder-bot",
    version=version,
    url="https://github.com/anoadragon453/matrix-reminder-bot",
    description="A matrix bot to remind you about things!",
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "matrix-nio[e2e]>=0.10.0",
        "Markdown>=3.1.1",
        "PyYAML>=5.1.2",
        "dateparser>=0.7.4",
        "readabledelta>=0.0.2",
        "apscheduler>=3.6.3",
        "pytz>=2020.1",
    ],
    extras_require={
        "postgres": ["psycopg2>=2.8.5"],
        "dev": [
            "isort==5.0.4",
            "flake8==3.8.3",
            "flake8-comprehensions==3.2.3",
            "black==19.10b0",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    # Allow the user to run the bot with `matrix-reminder-bot ...`
    scripts=["matrix-reminder-bot"],
)



import logging
from datetime import datetime, timedelta
from typing import Dict, Tuple

import pytz
from apscheduler.util import timedelta_seconds
from nio import AsyncClient

from matrix_reminder_bot.config import CONFIG
from matrix_reminder_bot.reminder import REMINDERS, Reminder

latest_migration_version = 3

logger = logging.getLogger(__name__)


class Storage(object):
    def __init__(self, client: AsyncClient):
        """Setup the database

        Runs an initial setup or migrations depending on whether a database file has already
        been created

        Args:
            client: The matrix client
        """
        # Check which type of database has been configured
        self.client = client
        self.conn = self._get_database_connection(
            CONFIG.database.type, CONFIG.database.connection_string
        )
        self.cursor = self.conn.cursor()
        self.db_type = CONFIG.database.type

        # Try to check the current migration version
        migration_level = 0
        try:
            self._execute("SELECT version FROM migration_version")
            row = self.cursor.fetchone()
            migration_level = row[0]
        except Exception:
            self._initial_db_setup()
        finally:
            if migration_level < latest_migration_version:
                self._run_db_migrations(migration_level)

        # Load reminders from the db
        REMINDERS.update(self._load_reminders())

        logger.info(f"Database initialization of type '{self.db_type}' complete")

    def _get_database_connection(self, database_type: str, connection_string: str):
        if database_type == "sqlite":
            import sqlite3

            # Initialize a connection to the database, with autocommit on
            return sqlite3.connect(connection_string, isolation_level=None)
        elif database_type == "postgres":
            import psycopg2

            conn = psycopg2.connect(connection_string)

            # Autocommit on
            conn.set_isolation_level(0)

            return conn

    def _execute(self, *args):
        """A wrapper around cursor.execute that transforms ?'s to %s for postgres"""
        if self.db_type == "postgres":
            self.cursor.execute(args[0].replace("?", "%s"), *args[1:])
        else:
            self.cursor.execute(*args)

    def _initial_db_setup(self):
        """Initial setup of the database"""
        logger.info("Performing initial database setup...")

        # Set up the migration_version table
        self._execute(
            """
            CREATE TABLE migration_version (
                version INTEGER PRIMARY KEY
            )
        """
        )

        # Initially set the migration version to 0
        self._execute(
            """
            INSERT INTO migration_version (
                version
            ) VALUES (?)
        """,
            (0,),
        )

        # Set up the reminders table
        self._execute(
            """
            CREATE TABLE reminder (
                text TEXT,
                start_time TEXT NOT NULL,
                recurse_timedelta_s INTEGER,
                room_id TEXT NOT NULL,
                target_user TEXT,
                alarm BOOL NOT NULL
            )
        """
        )

        # Create a unique index on room_id, reminder text as no two reminders in the same
        # room can have the same reminder text
        self._execute(
            """
            CREATE UNIQUE INDEX reminder_room_id_text
            ON reminder(room_id, text)
        """
        )

    def _run_db_migrations(self, current_migration_version: int):
        """Execute database migrations. Migrates the database to the
        `latest_migration_version`

        Args:
            current_migration_version: The migration version that the database is
                currently at
        """
        logger.debug("Checking for necessary database migrations...")

        if current_migration_version < 1:
            logger.info("Migrating the database from v0 to v1...")

            # Add cron_tab column, prevent start_time from being required
            #
            # As SQLite3 is quite limited, we need to create a new table and populate it
            # with existing data
            self._execute("ALTER TABLE reminder RENAME TO reminder_temp")

            self._execute(
                """
                CREATE TABLE reminder (
                    text TEXT,
                    start_time TEXT,
                    recurse_timedelta_s INTEGER,
                    cron_tab TEXT,
                    room_id TEXT NOT NULL,
                    target_user TEXT,
                    alarm BOOL NOT NULL
                )
           """
            )
            self._execute(
                """
                INSERT INTO reminder (
                    text,
                    start_time,
                    recurse_timedelta_s,
                    room_id,
                    target_user,
                    alarm
                )
                SELECT
                    text,
                    start_time,
                    recurse_timedelta_s,
                    room_id,
                    target_user,
                    alarm
                FROM reminder_temp;
           """
            )

            self._execute(
                """
                 DROP INDEX reminder_room_id_text
           """
            )
            self._execute(
                """
                CREATE UNIQUE INDEX reminder_room_id_text
                ON reminder(room_id, text)
           """
            )

            self._execute(
                """
                DROP TABLE reminder_temp
           """
            )

            self._execute(
                """
                 UPDATE migration_version SET version = 1
            """
            )

            logger.info("Database migrated to v1")

        if current_migration_version < 2:
            logger.info("Migrating the database from v1 to v2...")

            # Add a timezone column to the reminder database, so we can easily keep
            # track of which timezone a reminder was created in
            self._execute(
                """
                ALTER TABLE reminder
                    ADD COLUMN timezone TEXT
            """
            )

            # Assume the currently configured database timezone for all rows
            self._execute(
                """
                UPDATE reminder SET timezone = ?
            """,
                (CONFIG.timezone,),
            )

            self._execute(
                """
                 UPDATE migration_version SET version = 2
            """
            )

            logger.info("Database migrated to v2")

        if current_migration_version < 3:
            logger.info("Migrating the database from v2 to v3...")

            # Remove current timezone information from all start_time entries (as an older
            # version of the code used to insert them)
            self._execute(
                """
                SELECT text, room_id, start_time FROM reminder
                    WHERE start_time LIKE '%+%'
                    OR start_time LIKE '%-%'
            """
            )
            rows = self.cursor.fetchall()
            logger.debug("Loaded reminder rows with tz info: %s", rows)

            # Update start_time rows in the db with their non-timezone versions
            for row in rows:
                text = row[0]
                room_id = row[1]
                start_time = datetime.fromisoformat(row[2])

                # Remove timezone information from start_time
                start_time = start_time.replace(tzinfo=None)

                logger.debug(
                    "Updating (%s, %s) with new start_time: %s",
                    text,
                    room_id,
                    start_time,
                )
                self._execute(
                    """
                    UPDATE reminder SET start_time = ?
                        WHERE text = ? AND room_id = ?
                """,
                    (start_time, text, room_id),
                )

            self._execute(
                """
                 UPDATE migration_version SET version = 3
            """
            )

            logger.info("Database migrated to v3")

    def _load_reminders(self) -> Dict[Tuple[str, str], Reminder]:
        """Load reminders from the database

        Returns:
            A dictionary from (room_id, reminder text) to Reminder object
        """
        self._execute(
            """
            SELECT
                text,
                start_time,
                timezone,
                recurse_timedelta_s,
                cron_tab,
                room_id,
                target_user,
                alarm
            FROM reminder
        """
        )
        rows = self.cursor.fetchall()
        logger.debug("Loaded reminder rows: %s", rows)
        reminders = {}

        for row in rows:
            # Extract reminder data
            reminder_text = row[0]
            start_time = datetime.fromisoformat(row[1]) if row[1] else None
            timezone = row[2]
            recurse_timedelta = timedelta(seconds=row[3]) if row[3] else None
            cron_tab = row[4]
            room_id = row[5]
            target_user = row[6]
            alarm = row[7]

            if start_time:
                # If this is a one-off reminder whose start time is in the past, then it will
                # never fire. Ignore and delete the row from the db
                if not recurse_timedelta and not cron_tab:
                    now = datetime.now(tz=pytz.timezone(timezone))

                    # We don't replace the timezone in start_time itself as Reminder.__init__
                    # will add the timezone later (and doing so twice will produce strange
                    # behaviour)
                    if start_time.replace(tzinfo=pytz.timezone(timezone)) < now:
                        logger.debug(
                            "Deleting missed reminder in room %s: %s - %s",
                            room_id,
                            reminder_text,
                            start_time,
                        )

                        self.delete_reminder(room_id, reminder_text)
                        continue

            # Create and record the reminder
            reminders[(room_id, reminder_text.upper())] = Reminder(
                client=self.client,
                store=self,
                reminder_text=reminder_text,
                start_time=start_time,
                timezone=timezone,
                recurse_timedelta=recurse_timedelta,
                cron_tab=cron_tab,
                room_id=room_id,
                target_user=target_user,
                alarm=alarm,
            )

        return reminders

    def store_reminder(self, reminder: Reminder):
        """Store a new reminder in the database"""
        # timedelta.seconds does NOT give you the timedelta converted to seconds
        # Use a method from apscheduler instead
        if reminder.recurse_timedelta:
            delta_seconds = int(timedelta_seconds(reminder.recurse_timedelta))
        else:
            delta_seconds = None

        if reminder.start_time:
            # Remove timezone from start_time. We only want to store the timezone str
            # in the database
            reminder.start_time = reminder.start_time.replace(tzinfo=None)

        self._execute(
            """
            INSERT INTO reminder (
                text,
                start_time,
                timezone,
                recurse_timedelta_s,
                cron_tab,
                room_id,
                target_user,
                alarm
            ) VALUES (
                ?, ?, ?, ?, ?, ?, ?, ?
            )
        """,
            (
                reminder.reminder_text,
                reminder.start_time.isoformat() if reminder.start_time else None,
                reminder.timezone,
                delta_seconds,
                reminder.cron_tab,
                reminder.room_id,
                reminder.target_user,
                reminder.alarm,
            ),
        )

    def delete_reminder(self, room_id: str, reminder_text: str):
        """Delete a reminder via its reminder text and the room it was sent in"""
        self._execute(
            """
            DELETE FROM reminder WHERE room_id = ? AND text = ?
        """,
            (room_id, reminder_text),
        )
